from __future__ import annotations

import datetime as dt
import logging
import typing
from typing import Annotated, Any
from uuid import UUID

import inflection
from lapis_client_base import Absent
from pydantic import BaseModel, Field, Extra

from ..openapi import model as openapi

if typing.TYPE_CHECKING:
    from .elems.refs import ResolverFunc
    from .module_path import ModulePath

logger = logging.getLogger(__name__)

STRING_FORMATS = {
    'uuid': UUID,
    'date': dt.date,
    'date-time': dt.datetime,
}

PRIMITIVE_TYPES = {
    openapi.Type.string: str,
    openapi.Type.integer: int,
    openapi.Type.number: float,
    openapi.Type.boolean: bool,
}


def module_name(path: list[str]) -> str:
    return '.'.join([*path[:-1], inflection.underscore(path[-1])])


def get_type_ref(schema: openapi.Schema, module: ModulePath, name: str, required: bool, resolver: ResolverFunc) -> TypeRef:
    typ = _get_type_ref(schema, module, name, resolver)

    if schema.nullable:
        typ = typ.union_with(BuiltinTypeRef.from_str('None'))
    if not required:
        typ = typ.union_with(TypeRef.from_type(Absent))

    return typ


def _get_type_ref(schema: openapi.Schema, module: ModulePath, name: str, resolver: ResolverFunc) -> TypeRef:
    if schema.enum:
        return TypeRef(module=module.str(), name=name)
    elif schema.type == openapi.Type.string:
        return TypeRef.from_type(STRING_FORMATS.get(schema.format, str))
    elif schema.type in PRIMITIVE_TYPES:
        return BuiltinTypeRef.from_str(PRIMITIVE_TYPES[schema.type].__name__)
    elif schema.type == openapi.Type.object:
        return _get_type_ref_object(schema, module, name)
    elif schema.type == openapi.Type.array:
        return _get_type_ref_array(schema, module, name, resolver)
    return TypeRef.from_type(Any)


def _get_type_ref_object(schema: openapi.Schema, module: ModulePath, name: str) -> TypeRef:
    if schema.properties or schema.allOf:
        return TypeRef(module=module.str(), name=name)
    elif schema.anyOf:
        return BuiltinTypeRef.from_str('Unsupported')
    elif schema.oneOf:
        return BuiltinTypeRef.from_str('Unsupported')
    else:
        return TypeRef(module=module.str(), name=name)


def _get_type_ref_array(schema: openapi.Schema, module: ModulePath, parent_name: str, resolver: ResolverFunc) -> TypeRef:
    if isinstance(schema.items, openapi.Reference):
        item_schema, module, name = resolver(schema.items, openapi.Schema)
    else:
        item_schema = schema.items
        name = parent_name + 'Item'

    type_ref = get_type_ref(item_schema, module, name, True, resolver)
    return type_ref.list_of()


class TypeRef(BaseModel):
    module: str
    name: str

    class Config:
        extra = Extra.forbid

    def __repr__(self):
        return self.full_name()

    def full_name(self):
        return self.module + '.' + self.name if self.module != 'builtins' else self.name

    @staticmethod
    def from_str(path: str) -> TypeRef:
        module, name = path.rsplit('.', 1)
        return TypeRef(module=module, name=name)

    @staticmethod
    def from_type(typ: typing.Type) -> TypeRef:
        if hasattr(typ, '__origin__'):
            raise ValueError('Generic types unsupported', typ)
        module = typ.__module__
        name = typ.__name__
        if module == 'builtins':
            return BuiltinTypeRef.from_str(name)
        else:
            return TypeRef(module=module, name=name)

    def type_checking_imports(self) -> list[tuple[str, str]]:
        return []

    def imports(self) -> list[str]:
        return [self.module]

    def union_with(self, other: TypeRef) -> GenericTypeRef:
        return GenericTypeRef(module='typing', name='Union', args=[self, other])

    def list_of(self) -> GenericTypeRef:
        return GenericTypeRef(module='builtins', name='list', args=[self])

    def _types(self) -> list[TypeRef]:
        return [self]


class BuiltinTypeRef(TypeRef):
    module: str = 'builtins'
    schema_type = False

    class Config:
        extra = Extra.forbid

    def __repr__(self):
        return self.full_name()

    @staticmethod
    def from_str(name: str) -> BuiltinTypeRef:
        return BuiltinTypeRef(name=name)

    def imports(self) -> list[str]:
        return []

    def full_name(self):
        return self.name


class GenericTypeRef(TypeRef):
    schema_type = False
    args: Annotated[list[TypeRef], Field(default_factory=list)]

    class Config:
        extra = Extra.forbid

    def union_with(self, other: TypeRef) -> GenericTypeRef:
        if self.module == 'typing' and self.name == 'Union':
            return GenericTypeRef(module=self.module, name=self.name, args=[*self.args, other])
        else:
            return super().union_with(other)

    def type_checking_imports(self) -> list[tuple[str, str]]:
        return []

    def imports(self) -> list[str]:
        return [imp for typ in self._types() for imp in TypeRef.imports(typ)]

    def _types(self) -> list[TypeRef]:
        return [self, *[typ for arg in self.args for typ in arg._types()]]

    def __repr__(self) -> str:
        return self.full_name()

    # def full_name(self) -> str:
    #     return f'{super().full_name()}[{", ".join(arg.full_name() for arg in self.args)}]'



from dataclasses import dataclass, field
from typing import Union, Generator

from .operation_function import OperationFunctionModel, get_operation_func
from .refs import ResolverFunc
from ..module_path import ModulePath
from ...openapi import model as openapi


@dataclass(frozen=True)
class ClientClass:
    methods: list[OperationFunctionModel] = field(default_factory=list)


def get_client_class(openapi_model: openapi.OpenApiModel, module: ModulePath, resolver: ResolverFunc) -> ClientClass:
    functions = [
        get_operation_func(op, method, url_path, module / 'paths' / op.operationId, resolver)
        for url_path, path_item in openapi_model.paths.__root__.items()
        for method, op in get_operations(path_item, True)
    ]

    return ClientClass(
        methods=functions
    )


def get_operations(path_item: openapi.PathItem, skip_reference=False) -> Generator[tuple[str, Union[openapi.Operation, openapi.Reference]], None, None]:
    for op_name in path_item.__fields_set__:
        v = getattr(path_item, op_name)
        if (
                isinstance(v, openapi.Operation)
                or (not skip_reference and isinstance(v, openapi.Reference))
        ):
            yield op_name, v



import logging
from dataclasses import dataclass, field
from typing import Optional, Generator

import inflection

from .attribute import AttributeModel, get_attributes, get_enum_attribute
from .attribute_annotation import AttributeAnnotationModel
from .refs import ResolverFunc
from ..module_path import ModulePath
from ..type_ref import TypeRef, BuiltinTypeRef
from ...openapi import model as openapi

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SchemaClass:
    class_name: str
    base_type: TypeRef
    docstr: Optional[str] = None
    attributes: list[AttributeModel] = field(default_factory=list)


def get_schema_classes(
        schema: openapi.Schema,
        name: str,
        module: ModulePath,
        resolver: ResolverFunc,
) -> Generator[SchemaClass, None, None]:
    if schema.enum is not None:
        yield get_enum_class(schema, name)
        name = name + 'Value'

    schema_class = get_schema_class(schema, name, module, resolver)
    if schema_class is not None:
        yield schema_class

    # handle sub schemas

    if schema.type is openapi.Type.array:
        item_schema = schema.items
        if isinstance(item_schema, openapi.Schema):
            yield from get_schema_classes(item_schema, name + 'Item', module, resolver)
    elif schema.type is openapi.Type.object:
        if schema.properties:
            for prop_name, prop_schema in schema.properties.items():
                if not isinstance(prop_schema, openapi.Schema):
                    continue
                yield from get_schema_classes(prop_schema, name + inflection.camelize(prop_name), module, resolver)


def get_schema_class(
        schema: openapi.Schema,
        name: str,
        module: ModulePath,
        resolver: ResolverFunc,
) -> Optional[SchemaClass]:
    if schema.type is not openapi.Type.object:
        return None

    logger.debug(name)

    base_type = TypeRef.from_str('pydantic.BaseModel')
    attributes = get_attributes(schema, name, module, resolver) if schema.properties else []

    return SchemaClass(
        class_name=name,
        base_type=base_type,
        attributes=attributes,
        docstr=schema.description or None,
    )


def get_enum_class(
        schema: openapi.Schema,
        name: str
):
    return SchemaClass(
        class_name=name,
        base_type=TypeRef.from_str('enum.Enum'),
        attributes=[get_enum_attribute(v) for v in schema.enum],
        docstr=schema.description or None,
    )


def get_enum_value(value, schema: openapi.Schema) -> AttributeModel:
    # if schema.type == openapi.Type.string:
    if value is None:
        name = 'none'
    elif value == '':
        name = 'empty'
    else:
        name = value
    value = "'" + value.replace("'", r"\'") + "'" if value is not None else None
    return AttributeModel(name, AttributeAnnotationModel(BuiltinTypeRef(name='__ignored__'), dict(default=value)))



import logging
from pathlib import Path

import typer

from lapis.openapi.model import OpenApiModel
from lapis.render import render_client

logging.basicConfig()
logger = logging.getLogger('lapis')
logger.setLevel(logging.INFO)


def main(
        schema_path: Path,
        target_directory: Path,
        package_name: str,
):
    logger.info('Load schema')
    doc = load_schema(schema_path)
    logger.info('Parse schema')
    tree = OpenApiModel(**doc)
    render_client(tree, target_directory, package_name)


def load_schema(schema_path: Path):
    import pickle
    cache_path = schema_path.with_suffix('.pickle')
    if cache_path.exists():
        with open(cache_path, 'br') as fb:
            return pickle.load(fb)

    with open(schema_path, 'rt') as f:
        import yaml
        doc = yaml.safe_load(f)

    with open(cache_path, 'wb') as fb:
        pickle.dump(doc, fb)
    return doc


if __name__ == '__main__':
    typer.run(main)



import logging
from importlib import resources
from pathlib import Path

import tomlkit
from pydantic import BaseModel, Extra

from ...openapi import model as openapi

logger = logging.getLogger(__name__)


class PyProject(BaseModel):
    name: str
    description: str
    version: str
    authors: list[str] = []

    class Config:
        extra = Extra.forbid


def get_pyproject(info: openapi.Info) -> PyProject:
    return PyProject(
        name=info.title,
        description=info.description or "",
        version=info.version,
    )


def render_pyproject(root: Path, model: PyProject) -> None:
    target = root / 'pyproject.toml'
    if target.exists():
        logger.info('pyproject.toml exists, skipping')
        return

    with resources.open_text('lapis.render.templates', 'pyproject.toml') as fbuf:
        d: dict = tomlkit.loads(fbuf.read())

    d.setdefault('tool', {}).setdefault('poetry', {}).update(model.dict())

    with open(target, 'tw') as fbuf:
        fbuf.write(tomlkit.dumps(d, True))



from dataclasses import dataclass, field
from typing import Union, Optional

import inflection

from .module import AbstractModule, template_imports
from .param_model_class import get_param_model_classes
from .refs import ResolverFunc
from .schema_class import SchemaClass, get_schema_classes
from ..module_path import ModulePath
from ...openapi import model as openapi


@dataclass(frozen=True)
class SchemaModule(AbstractModule):
    """
    One schema module per schema element directly under #/components/schemas, containing that schema and all non-reference schemas.
    One schema module for all schemas under each operation

    """
    body: list[SchemaClass] = field(default_factory=list)


def get_modules_for_components_schemas(
        schemas: dict[str, Union[openapi.Schema, openapi.Reference]], root_package: ModulePath, resolver: ResolverFunc
) -> list[SchemaModule]:
    result = []
    for name, schema in schemas.items():
        if isinstance(schema, openapi.Schema):
            module = get_module_for_components_schema(schema, root_package / inflection.underscore(name), resolver)
            if module is not None:
                result.append(module)
    return result


def get_module_for_components_schema(schema: openapi.Schema, path: ModulePath, resolver: ResolverFunc) -> Optional[SchemaModule]:
    classes = [cls for cls in get_schema_classes(schema, inflection.camelize(path.parts[-1]), path, resolver)]
    if len(classes):
        return _get_module_for_components_schema(classes, path)


def _get_module_for_components_schema(classes: list[SchemaClass], path: ModulePath) -> SchemaModule:
    imports = list({
        imp
        for cls in classes
        if cls.base_type is not None
        for imp in cls.base_type.imports()
        if imp not in template_imports
    })

    type_checking_imports = list({
        import_
        for schema_class in classes
        for attr in schema_class.attributes
        for import_ in attr.annotation.type.imports()
        if import_ not in imports and import_ not in template_imports and import_ != path.str()
    })
    type_checking_imports.sort()

    return SchemaModule(
        path=path,
        body=classes,
        imports=imports,
        type_checking_imports=type_checking_imports,
    )


def get_module_for_param_model_classes(op: openapi.Operation, module: ModulePath, resolve: ResolverFunc) -> SchemaModule:
    mod = module / 'schemas'
    classes = [cls for cls in get_param_model_classes(op, op.operationId, mod, resolve)]
    return _get_module_for_components_schema(classes, mod)



from dataclasses import dataclass
from typing import Optional

from lapis_client_base import ParamDirection


@dataclass(frozen=True)
class ArgModel:
    name: str
    annotation: str

    deprecated = False
    style: Optional[str] = None
    explode: Optional[bool] = None
    direction: ParamDirection = ParamDirection.read | ParamDirection.write
    allowReserved: Optional[bool] = False



from typing import Generator

import inflection

from .attribute import AttributeModel
from .attribute_annotation import get_attr_annotation
from .refs import ResolverFunc
from .schema_class import SchemaClass, get_schema_classes
from ..module_path import ModulePath
from ..type_ref import TypeRef
from ...openapi import model as openapi

"""
Param model is a synthetic (from the perspective of OpenAPI specification) object that holds and validates all Operation parameters.
"""


def get_param_attribute(
        param: openapi.Parameter,
        parent_name: str,
        module: ModulePath,
        resolver: ResolverFunc,
) -> AttributeModel:
    return AttributeModel(
        name=param.in_[0] + '_' + param.name.replace(':', '_'),
        annotation=get_attr_annotation(param.schema_, param.name, parent_name, param.required, module, resolver, param.in_),
        deprecated=param.deprecated,
    )


def get_param_model_classes(
        operation: openapi.Operation,
        name: str,
        module: ModulePath,
        resolver: ResolverFunc,
) -> Generator[SchemaClass, None, None]:
    schema_class = get_param_model_class(operation, name, module, resolver)
    if schema_class is not None:
        yield schema_class

    # handle sub schemas
    for param in operation.parameters:
        schema = param.schema_
        if not isinstance(schema, openapi.Schema):
            continue
        yield from get_schema_classes(schema, inflection.camelize(name) + inflection.camelize(param.name), module, resolver)


def get_param_model_class(
        operation: openapi.Operation,
        name: str,
        module: ModulePath,
        resolver: ResolverFunc,
) -> SchemaClass:
    attributes = [get_param_attribute(param, name, module, resolver) for param in
                  operation.parameters] if operation.parameters else []
    base_type = TypeRef.from_str('pydantic.BaseModel')

    return SchemaClass(
        class_name=inflection.camelize(name),
        base_type=base_type,
        attributes=attributes,
        docstr=operation.description or None,
    )



from dataclasses import dataclass
from typing import Optional, Any

import inflection

from .refs import SchemaOrRef, ResolverFunc
from ..module_path import ModulePath
from ..type_ref import TypeRef, get_type_ref
from ...openapi import model as openapi


@dataclass(frozen=True)
class AttributeAnnotationModel:
    type: TypeRef
    field_props: dict[str, Any]

    direction: Optional[str] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = False


def get_attr_annotation(
        typ: SchemaOrRef,
        name: str,
        parent_name: str,
        required: bool,
        module: ModulePath,
        resolve: ResolverFunc,
        in_: Optional[str] = None,
) -> AttributeAnnotationModel:
    """
    if typ is a schema, then it's a nested schema. Name should be parent_class_name+prop_name, and module is the same.
    Otherwise, it's a reference, schema, module and name should be resolved from it and used to generate type_ref
    """
    if isinstance(typ, openapi.Reference):
        schema, module, name = resolve(typ, openapi.Schema)
    else:
        schema: openapi.Schema = typ
        name = inflection.camelize(parent_name) + inflection.camelize(name)
    return _get_attr_annotation(schema, name, required, module, resolve, in_)


FIELD_PROPS = {
    'multipleOf': 'multiple_of',
    'maximum': 'le',
    'exclusiveMaximum': 'lt',
    'minimum': 'gt',
    'exclusiveMinimum': 'ge',
    'maxLength': 'max_length',
    'minLength': 'min_length',
    'maxItems': 'max_items',
    'minItems': 'min_items',
    'uniqueItems': 'unique_items',
    'maxProperties': 'max_properties',
    'minProperties': 'min_properties',
    'default': 'default',
}


def _get_attr_annotation(
        schema: openapi.Schema,
        name: str,
        required: bool,
        module: ModulePath,
        resolve: ResolverFunc,
        in_: Optional[str] = None,
) -> AttributeAnnotationModel:
    field_props = {FIELD_PROPS[k]: getattr(schema, k) for k in schema.__fields_set__ if k in FIELD_PROPS}
    for k, v in field_props.items():
        if isinstance(v, str):
            field_props[k] = f"'{v}'"

    if in_ is not None:
        field_props['in_'] = 'lapis_client_base.ParamPlacement.' + in_

    if 'pattern' in schema.__fields_set__:
        field_props['regex'] = f"r'${schema.pattern}'"

    if not required:
        field_props['default'] = 'lapis_client_base.absent.ABSENT'

    return AttributeAnnotationModel(
        type=get_type_ref(schema, module, name, required, resolve),
        direction=get_direction(schema.readOnly, schema.writeOnly),
        field_props=field_props
    )


def get_direction(read_only: Optional[bool], write_only: Optional[bool]) -> Optional[str]:
    if read_only:
        if write_only:
            raise ValueError()
        else:
            return 'lapis_client_base.ParamDirection.read'
    else:
        if write_only:
            return 'lapis_client_base.ParamDirection.write'
        else:
            return None



from dataclasses import dataclass

from .client_class import ClientClass, get_client_class
from .module import AbstractModule, template_imports
from .refs import ResolverFunc
from ..module_path import ModulePath
from ...openapi import model as openapi


@dataclass(frozen=True, kw_only=True)
class ClientModule(AbstractModule):
    body: ClientClass


def get_client_class_module(model: openapi.OpenApiModel, client_module: ModulePath, root_module: ModulePath, resolver: ResolverFunc) -> ClientModule:
    client_class = get_client_class(model, root_module, resolver)

    imports = [
        'lapis_client_base',
    ]

    type_checking_imports = list({
        imp
        for attr in client_class.methods
        for t in attr.params
        for imp in t.annotation.type.imports()
        if imp not in imports and imp not in template_imports
    })
    type_checking_imports.sort()

    return ClientModule(
        path=client_module,
        imports=imports,
        type_checking_imports=type_checking_imports,
        body=client_class,
    )



"""
Schema mapping
as class:

    title -  docstr
    maxItems: config - validation
    minItems: config - validation
    uniqueItems: config - validation
    maxProperties: config - validation
    minProperties: config - validation
    required: child attributes
    enum: enum class
    type: discriminator
    not_: config - validation
    allOf: superclasses
    oneOf: typing.Union
    anyOf: superclasses, all fields non-required, must validate against at least one superclass
    items: ignore this schema and generate class for the one declared here
    properties: attributes
    additionalProperties: unsupported
    description: docstr
    format: unused
    default: unsupported
    nullable: unused
    discriminator: unsupported
    readOnly: unused
    writeOnly: unused
    example: unused
    externalDocs: docstr
    deprecated: unused
    xml: unsupported


as attribute
    multipleOf: Field property
    maximum: Field property
    exclusiveMaximum: Field property
    minimum: Field property
    exclusiveMinimum: Field property
    maxLength: Field property
    minLength: Field property
    pattern: Field property
    maxItems: unused
    minItems: unused
    uniqueItems: unused
    maxProperties: unused
    minProperties: unused
    required: unused directly (passed as boolean)
    enum: unused
    type: type discriminator
    additionalProperties: unsupported
    format: type hint
    default: unsupported
    nullable: make Union[None, this]
    readOnly: Field property
    writeOnly: Field property
    deprecated: unsupported
    xml: unsupported


"""



import logging
import os
from pathlib import Path

from jinja2 import Environment, PackageLoader

from .black import format_code
from .elems.client_class import get_operations
from .elems.client_module import get_client_class_module
from .elems.module import AbstractModule
from .elems.pyproject import get_pyproject, render_pyproject
from .elems.refs import get_resolver, ResolverFunc
from .elems.schema_module import get_modules_for_components_schemas, get_module_for_param_model_classes
from .module_path import ModulePath
from ..openapi import model as openapi

logger = logging.getLogger(__name__)


def render(source: str, destination: Path, render_model: AbstractModule, env: Environment):
    try:
        destination.parent.mkdir(parents=True, exist_ok=True)
        text = env.get_template(source).render(model=render_model)
        text = format_code(text)
        with open(destination, 'wt') as fb:
            fb.write(text)
    except Exception:
        print(render_model)
        raise


def render_client(model: openapi.OpenApiModel, target: Path, package_name: str) -> None:
    env = Environment(
        keep_trailing_newline=True,
        loader=PackageLoader("lapis.render"),
    )

    render_pyproject(target, get_pyproject(model.info))

    gen_root = target / 'gen'
    gen_root.mkdir(parents=True, exist_ok=True)

    resolver = get_resolver(model, package_name)

    render_client_module(model, package_name, gen_root, resolver, env)
    render_schema_modules(model, package_name, gen_root, resolver, env)

    ensure_init_py(gen_root, package_name)


def render_client_module(model: openapi.OpenApiModel, package_name: str, gen_root: Path, resolver: ResolverFunc, env: Environment):
    logger.info('Render client modules')
    root_mod = ModulePath(package_name)
    client_module = root_mod / 'client.py'
    client_class_module = get_client_class_module(model, client_module, root_mod, resolver)
    path = client_module.to_path(gen_root)
    render('client_class.py.jinja2', path, client_class_module, env)


def render_schema_modules(model: openapi.OpenApiModel, package_name: str, gen_root: Path, resolver: ResolverFunc, env: Environment):
    root_module = ModulePath(package_name)

    if model.components.schemas:
        logger.info('Render schema modules')
        modules = get_modules_for_components_schemas(model.components.schemas, root_module / 'components' / 'schemas', resolver)
        for module in modules:
            render('schema_class.py.jinja2', module.path.to_path(gen_root).with_suffix('.py'), module, env)

    for path, path_item in model.paths.__root__.items():
        for tpl in get_operations(path_item, True):
            method, op = tpl
            if op.parameters:
                module_path = root_module / 'paths' / op.operationId / 'schemas'
                mod = get_module_for_param_model_classes(op, module_path, resolver)
                render('schema_class.py.jinja2', module_path.to_path(gen_root).with_suffix('.py'), mod, env)


def ensure_init_py(gen_root, package_name):
    for (dirpath, dirnames, filenames) in os.walk(gen_root / package_name):
        if '__init__.py' not in filenames:
            (Path(dirpath) / '__init__.py').touch()



# generated by datamodel-codegen:
#   filename:  schema.yaml

from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Dict, List, Optional, Union, Mapping

from pydantic import AnyUrl, BaseModel, EmailStr, Extra, Field


class Reference(BaseModel):
    class Config:
        extra = Extra.forbid

    ref: Annotated[str, Field(alias='$ref')]


class Contact(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[str]
    url: Optional[str]
    email: Optional[EmailStr]


class License(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str
    url: Optional[str]


class ServerVariable(BaseModel):
    class Config:
        extra = Extra.forbid

    enum: Optional[List[str]]
    default: str
    description: Optional[str]


class Type(Enum):
    array = 'array'
    boolean = 'boolean'
    integer = 'integer'
    number = 'number'
    object = 'object'
    string = 'string'


class Discriminator(BaseModel):
    propertyName: str
    mapping: Optional[Dict[str, str]]


class XML(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[str]
    namespace: Optional[AnyUrl]
    prefix: Optional[str]
    attribute: Optional[bool] = False
    wrapped: Optional[bool] = False


class Example(BaseModel):
    class Config:
        extra = Extra.forbid

    summary: Optional[str]
    description: Optional[str]
    value: Optional[Any]
    externalValue: Optional[str]


class Style(Enum):
    simple = 'simple'


class Paths(BaseModel):
    __root__: Annotated[dict[str, Union[PathItem, Reference]], Field(default_factory=dict)]


class SecurityRequirement(BaseModel):
    __root__: Annotated[dict[str, list[str]], Field(default_factory=dict)]


class ExternalDocumentation(BaseModel):
    class Config:
        extra = Extra.forbid

    description: Optional[str]
    url: str


class ExampleXORExamples(BaseModel):
    __root__: Annotated[
        Any,
        Field(
            description='Example and examples are mutually exclusive',
            not_={'required': ['example', 'examples']},
        ),
    ]


class SchemaXORContentItem(Reference):
    """
    Some properties are not allowed if content is present
    """

    pass


class SchemaXORContent(BaseModel):
    __root__: Annotated[
        Union[Any, SchemaXORContentItem],
        Field(
            description='Schema and content are mutually exclusive, at least one is required',
            not_={'required': ['schema', 'content']},
        ),
    ]


class In(Enum):
    path = 'path'


class Style1(Enum):
    matrix = 'matrix'
    label = 'label'
    simple = 'simple'


class Required(Enum):
    bool_True = True


class ParameterLocationItem(BaseModel):
    """
    Parameter in path
    """

    in_: Annotated[Optional[In], Field(alias='in')]
    style: Optional[Style1] = 'simple'
    required: Required


class In1(Enum):
    query = 'query'


class Style2(Enum):
    form = 'form'
    spaceDelimited = 'spaceDelimited'
    pipeDelimited = 'pipeDelimited'
    deepObject = 'deepObject'


class ParameterLocationItem1(BaseModel):
    """
    Parameter in query
    """

    in_: Annotated[Optional[In1], Field(alias='in')]
    style: Optional[Style2] = 'form'


class In2(Enum):
    header = 'header'


class ParameterLocationItem2(BaseModel):
    """
    Parameter in header
    """

    in_: Annotated[Optional[In2], Field(alias='in')]
    style: Optional[Style] = 'simple'


class In3(Enum):
    cookie = 'cookie'


class Style4(Enum):
    form = 'form'


class ParameterLocationItem3(BaseModel):
    """
    Parameter in cookie
    """

    in_: Annotated[Optional[In3], Field(alias='in')]
    style: Optional[Style4] = 'form'


class ParameterLocation(BaseModel):
    __root__: Annotated[
        Union[
            ParameterLocationItem,
            ParameterLocationItem1,
            ParameterLocationItem2,
            ParameterLocationItem3,
        ],
        Field(description='Parameter location'),
    ]


class Type1(Enum):
    apiKey = 'apiKey'


class In4(Enum):
    header = 'header'
    query = 'query'
    cookie = 'cookie'


class APIKeySecurityScheme(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type1
    name: str
    in_: Annotated[In4, Field(alias='in')]
    description: Optional[str]


class Type2(Enum):
    http = 'http'


class HTTPSecurityScheme1(BaseModel):
    class Config:
        extra = Extra.forbid

    scheme: str
    bearerFormat: Optional[str]
    description: Optional[str]
    type: Type2


class Scheme(Enum):
    bearer = 'bearer'


class HTTPSecuritySchemeItem(BaseModel):
    """
    Bearer
    """

    scheme: Optional[Scheme]


class HTTPSecuritySchemeItem1(BaseModel):
    """
    Non Bearer
    """

    scheme: Annotated[Optional[Any], Field(not_={'enum': ['bearer']})]


class HTTPSecurityScheme(BaseModel):
    class Config:
        extra = Extra.forbid

    __root__: Union[
        HTTPSecurityScheme1, HTTPSecuritySchemeItem, HTTPSecuritySchemeItem1
    ]


class Type3(Enum):
    oauth2 = 'oauth2'


class Type4(Enum):
    openIdConnect = 'openIdConnect'


class OpenIdConnectSecurityScheme(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type4
    openIdConnectUrl: str
    description: Optional[str]


class ImplicitOAuthFlow(BaseModel):
    class Config:
        extra = Extra.forbid

    authorizationUrl: str
    refreshUrl: Optional[str]
    scopes: Dict[str, str]


class PasswordOAuthFlow(BaseModel):
    class Config:
        extra = Extra.forbid

    tokenUrl: str
    refreshUrl: Optional[str]
    scopes: Optional[Dict[str, str]]


class ClientCredentialsFlow(PasswordOAuthFlow):
    pass


class AuthorizationCodeOAuthFlow(BaseModel):
    class Config:
        extra = Extra.forbid

    authorizationUrl: str
    tokenUrl: str
    refreshUrl: Optional[str]
    scopes: Optional[Dict[str, str]]


class Callback(BaseModel):
    __root__: Annotated[dict[str, Union[PathItem, Reference]], Field(default_factory=dict)]


class Info(BaseModel):
    class Config:
        extra = Extra.forbid

    title: str
    description: Optional[str]
    termsOfService: Optional[str]
    contact: Optional[Contact]
    license: Optional[License]
    version: str


class Server(BaseModel):
    class Config:
        extra = Extra.forbid

    url: str
    description: Optional[str]
    variables: Optional[Dict[str, ServerVariable]]


class Schema(BaseModel):
    class Config:
        extra = Extra.forbid

    title: Optional[str]
    multipleOf: Annotated[Optional[float], Field(gt=0.0)]
    maximum: Optional[float]
    exclusiveMaximum: Optional[bool] = False
    minimum: Optional[float]
    exclusiveMinimum: Optional[bool] = False
    maxLength: Annotated[Optional[int], Field(ge=0)]
    minLength: Annotated[Optional[int], Field(ge=0)] = 0
    pattern: Optional[str]
    maxItems: Annotated[Optional[int], Field(ge=0)]
    minItems: Annotated[Optional[int], Field(ge=0)] = 0
    uniqueItems: Optional[bool] = False
    maxProperties: Annotated[Optional[int], Field(ge=0)]
    minProperties: Annotated[Optional[int], Field(ge=0)] = 0
    required: Annotated[Optional[List[str]], Field(min_items=1, unique_items=True)]
    enum: Annotated[Optional[List], Field(min_items=1, unique_items=False)]
    type: Optional[Type]
    not_: Annotated[Optional[Union[Schema, Reference]], Field(alias='not')]
    allOf: Optional[List[Union[Schema, Reference]]]
    oneOf: Optional[List[Union[Schema, Reference]]]
    anyOf: Optional[List[Union[Schema, Reference]]]
    items: Optional[Union[Schema, Reference]]
    properties: Optional[Dict[str, Union[Schema, Reference]]]
    additionalProperties: Optional[Union[Schema, Reference, bool]] = True
    description: Optional[str]
    format: Optional[str]
    default: Optional[Any]
    nullable: Optional[bool] = False
    discriminator: Optional[Discriminator]
    readOnly: Optional[bool] = False
    writeOnly: Optional[bool] = False
    example: Optional[Any]
    externalDocs: Optional[ExternalDocumentation]
    deprecated: Optional[bool] = False
    xml: Optional[XML]


class Tag(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str
    description: Optional[str]
    externalDocs: Optional[ExternalDocumentation]


class OAuthFlows(BaseModel):
    class Config:
        extra = Extra.forbid

    implicit: Optional[ImplicitOAuthFlow]
    password: Optional[PasswordOAuthFlow]
    clientCredentials: Optional[ClientCredentialsFlow]
    authorizationCode: Optional[AuthorizationCodeOAuthFlow]


class Link(BaseModel):
    class Config:
        extra = Extra.forbid

    operationId: Optional[str]
    operationRef: Optional[str]
    parameters: Optional[Dict[str, Any]]
    requestBody: Optional[Any]
    description: Optional[str]
    server: Optional[Server]


class OAuth2SecurityScheme(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type3
    flows: OAuthFlows
    description: Optional[str]


class SecurityScheme(BaseModel):
    __root__: Union[
        APIKeySecurityScheme,
        HTTPSecurityScheme,
        OAuth2SecurityScheme,
        OpenIdConnectSecurityScheme,
    ]


def ref_to_path(ref: str) -> list[str]:
    return ref.split('/')[1:]


class OpenApiModel(BaseModel):
    """
    Validation schema for OpenAPI Specification 3.0.X.
    """

    class Config:
        extra = Extra.forbid

    openapi: Annotated[str, Field(regex='^3\\.0\\.\\d(-.+)?$')]
    info: Info
    externalDocs: Optional[ExternalDocumentation]
    servers: Optional[List[Server]]
    security: Optional[List[SecurityRequirement]]
    tags: Annotated[Optional[List[Tag]], Field(unique_items=True)]
    paths: Paths
    components: Optional[Components]

    def resolve(self, ref: str, t: Optional[type]) -> Any:
        result = self._schema_get(self.recursive_resolve(ref))
        if t is not None and not isinstance(result, t):
            raise TypeError(ref, t, type(result))
        else:
            return result

    def recursive_resolve(self, ref: str) -> str:
        """
        Resolve recursive references
        :return: The last reference, which points to a non-reference object
        """

        stack = []

        while True:
            model = self._schema_get(ref)
            if isinstance(model, Reference):
                ref = model.ref
                if ref in stack:
                    raise RecursionError(stack, ref)
                else:
                    stack.append(ref)
            else:
                return ref

    def _schema_get(self, ref: str) -> Any:
        model = self
        path = ref_to_path(ref)
        for part in path:
            model = model[part] if isinstance(model, Mapping) else getattr(model, part)
        return model


class Components(BaseModel):
    class Config:
        extra = Extra.forbid

    schemas: Optional[Dict[str, Union[Schema, Reference]]]
    responses: Optional[Dict[str, Union[Reference, Response]]]
    parameters: Optional[Dict[str, Union[Reference, Parameter]]]
    examples: Optional[Dict[str, Union[Reference, Example]]]
    requestBodies: Optional[Dict[str, Union[Reference, RequestBody]]]
    headers: Optional[Dict[str, Union[Reference, Header]]]
    securitySchemes: Optional[Dict[str, Union[Reference, SecurityScheme]]]
    links: Optional[Dict[str, Union[Reference, Link]]]
    callbacks: Optional[Dict[str, Union[Reference, Callback]]]


class Response(BaseModel):
    class Config:
        extra = Extra.forbid

    description: str
    headers: Optional[Dict[str, Union[Header, Reference]]]
    content: Optional[Dict[str, MediaType]]
    links: Optional[Dict[str, Union[Link, Reference]]]


class MediaType(BaseModel):
    class Config:
        extra = Extra.forbid

    schema_: Annotated[Optional[Union[Schema, Reference]], Field(alias='schema')]
    example: Optional[Any]
    examples: Optional[Dict[str, Union[Example, Reference]]]
    encoding: Optional[Dict[str, Encoding]]


class Header(BaseModel):
    class Config:
        extra = Extra.forbid

    description: Optional[str]
    required: Optional[bool] = False
    deprecated: Optional[bool] = False
    allowEmptyValue: Optional[bool] = False
    style: Optional[Style] = 'simple'
    explode: Optional[bool]
    allowReserved: Optional[bool] = False
    schema_: Annotated[Optional[Union[Schema, Reference]], Field(alias='schema')]
    content: Annotated[
        Optional[Dict[str, MediaType]], Field(maxProperties=1, minProperties=1)
    ]
    example: Optional[Any]
    examples: Optional[Dict[str, Union[Example, Reference]]]


class PathItem(BaseModel):
    class Config:
        extra = Extra.forbid

    summary: Optional[str]
    description: Optional[str]
    servers: Optional[List[Server]]
    parameters: Annotated[
        Optional[List[Union[Parameter, Reference]]], Field(unique_items=True)
    ]
    get: Optional[Operation]
    put: Optional[Operation]
    post: Optional[Operation]
    delete: Optional[Operation]
    options: Optional[Operation]
    head: Optional[Operation]
    patch: Optional[Operation]
    trace: Optional[Operation]


class Operation(BaseModel):
    class Config:
        extra = Extra.forbid

    tags: Optional[List[str]]
    summary: Optional[str]
    description: Optional[str]
    externalDocs: Optional[ExternalDocumentation]
    operationId: Optional[str]
    parameters: Annotated[
        Optional[List[Union[Parameter, Reference]]], Field(unique_items=True)
    ]
    requestBody: Optional[Union[RequestBody, Reference]]
    responses: Responses
    callbacks: Optional[Dict[str, Union[Callback, Reference]]]
    deprecated: Optional[bool] = False
    security: Optional[List[SecurityRequirement]]
    servers: Optional[List[Server]]


class Responses(BaseModel):
    __root__: Annotated[Optional[dict[str, Union[Response, Reference]]], Field(default_factory=dict)]


class Parameter(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str
    in_: Annotated[str, Field(alias='in')]
    description: Optional[str]
    required: Optional[bool] = False
    deprecated: Optional[bool] = False
    allowEmptyValue: Optional[bool] = False
    style: Optional[str]
    explode: Optional[bool]
    allowReserved: Optional[bool] = False
    schema_: Annotated[Optional[Union[Schema, Reference]], Field(alias='schema')]
    content: Annotated[
        Optional[Dict[str, MediaType]], Field(maxProperties=1, minProperties=1)
    ]
    example: Optional[Any]
    examples: Optional[Dict[str, Union[Example, Reference]]]


class RequestBody(BaseModel):
    class Config:
        extra = Extra.forbid

    description: Optional[str]
    content: Dict[str, MediaType]
    required: Optional[bool] = False


class Encoding(BaseModel):
    class Config:
        extra = Extra.forbid

    contentType: Optional[str]
    headers: Optional[Dict[str, Header]]
    style: Optional[Style2]
    explode: Optional[bool]
    allowReserved: Optional[bool] = False


Schema.update_forward_refs()
OpenApiModel.update_forward_refs()
Components.update_forward_refs()
Response.update_forward_refs()
MediaType.update_forward_refs()
PathItem.update_forward_refs()
Operation.update_forward_refs()
Paths.update_forward_refs()



from __future__ import annotations

import pathlib
from typing import Union, Iterable


class ModulePath:
    _SEP = '.'

    def __init__(self, module: Union[str, list[str]]):
        self.parts = module if isinstance(module, list) else module.split(ModulePath._SEP)

    def str(self):
        return ModulePath._SEP.join(self.parts)

    def to_path(self, root: pathlib.Path):
        return root.joinpath(*self.parts)

    def __truediv__(self, other: Union[str, Iterable[str]]):
        if isinstance(other, str):
            other = [other]
        return ModulePath([*self.parts, *other])

    def __repr__(self):
        return self.str()

    def __eq__(self, other: ModulePath):
        return self.parts == other.parts



import functools
import logging
from typing import Callable, TypeVar, Union, TypeAlias, Type, cast

import inflection

from ..module_path import ModulePath
from ...openapi import model as openapi

logger = logging.getLogger(__name__)

T = TypeVar('T', openapi.Schema, openapi.Parameter, openapi.SecurityScheme)
ResolverFunc = Callable[[openapi.Reference, Type[T]], tuple[T, ModulePath, str]]

SchemaOrRef: TypeAlias = Union[openapi.Schema, openapi.Reference]


def resolve(model: openapi.OpenApiModel, root_package: str, ref: openapi.Reference, typ: Type[T]) -> tuple[T, ModulePath, str]:
    """
    module = {root_package}.{path[0:4]}
    name = path[4:]
    """

    path = openapi.ref_to_path(model.recursive_resolve(ref.ref))

    if path[0] == 'paths':
        op = model.resolve(_mkref(path[:4]), openapi.Operation)
        if op.operationId:
            path[2:4] = op.operationId

    module = ModulePath(root_package) / path[:-1] / inflection.underscore(path[-1])
    return model.resolve(_mkref(path), typ), module, path[-1]


def get_resolver(model: openapi.OpenApiModel, package: str) -> ResolverFunc:
    return cast(ResolverFunc, functools.partial(resolve, model, package))


def _mkref(s: list[str]) -> str:
    return '/'.join(['#', *s])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['lapis', 'lapis.openapi', 'lapis.render', 'lapis.render.elems']

package_data = \
{'': ['*'], 'lapis.render': ['templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'black>=22.8.0,<23.0.0',
 'inflection>=0.5.1,<0.6.0',
 'lapis-client-base>=0.1.0,<0.2.0',
 'pydantic[email]>=1.9.2,<2.0.0',
 'tomlkit>=0.11.4,<0.12.0',
 'typer>=0.6.1,<0.7.0']

setup_kwargs = {
    'name': 'lapis-gen',
    'version': '0.1.1',
    'description': 'Python API client generator',
    'long_description': "# Lapis - Python OpenAPI client generator\n\nGenerate OpenAPI 3.0.3 client code that is easy to understand and debug.\n\nLeverages [Pydantic](https://github.com/pydantic/pydantic) as the base classes\nand [httpx](https://github.com/encode/httpx) as the HTTP client.\n\n## Supported OpenAPI features\n\n- Parameter names: operation parameters are uniquely identified by their name the value of an `in` attribute. It is possible to have parameter named `param` in all of path, query, cookies and headers.\n  \n  Lapis uses Hungarian notation for method parameter names.\n- Enums: [TODO] there's no limitation that enum schema cannot be an object or an array.\n\n  Enums might need two python classes - a subclass of `enum.Enum` and the schema class.\n- OneOf: [TODO] maps to typing.Union\n- AllOf: [TODO] maps to a separate class that uses all the schemas as superclasses.\n- AnyOf: [TODO] maps to similar class as in case of AllOf, all fields should be non-required and the object should validate against at least one of superclasses.\n- Recursive references between schemas: supported.\n- References to other schemas: unsupported.\n- Read- and write-only attributes: [TODO] Read-only attributes are considered non-existent when the object is validated before being sent to the server.\n\n## Broken and incomplete API specifications\n\nTODO: pre- and postprocessing of API specification (dict- and pydantic-based models) with python code, errata (jsonpatch), etc\n\n## Backwards compatibility\n\nOnce stable, Lapis should generate code that is backwards compatible as long as the API specification is too. The following rules are used to ensure that.\n\nNames of interface elements (public functions, classes and methods) are fully deterministic and are derived only from their source elements and, in some cases, their parents.\nFor example, schemas of the same name are either placed in separate modules, or their names have the parent element name prepended. \n\nEach operation must have operationId which is used to create a unique package for its models.\n\nThe structure of the API specification is roughly reflected in the generated code.\nEach schema under #/components/schemas together with all inline parameter schemas will become a single module in `<your_package>.components.schemas`.\nEach operation must define `operationId` which will be used as a sub-package name for the module containing its all inline schemas. Operation packages will be placed under `<your_package>.paths`\n\n",
    'author': 'Raphael Krupinski',
    'author_email': 'rafalkrupinski@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import re
from dataclasses import dataclass
from typing import Any

from .attribute_annotation import AttributeAnnotationModel, get_attr_annotation
from .refs import ResolverFunc, SchemaOrRef
from ..module_path import ModulePath
from ..type_ref import BuiltinTypeRef
from ...openapi import model as openapi

PYTHON_KEYWORDS = [
    'False',
    'None',
    'True',
    'and',
    'as',
    'assert',
    'break',
    'class',
    'continue',
    'def',
    'del',
    'elif',
    'else',
    'except',
    'finally',
    'for',
    'from',
    'global',
    'if',
    'import',
    'in',
    'is',
    'lambda',
    'nonlocal',
    'not',
    'or',
    'pass',
    'raise',
    'return',
    'try',
    'while',
    'with',
    'yield',
]


@dataclass(frozen=True)
class AttributeModel:
    name: str
    annotation: AttributeAnnotationModel
    deprecated: bool = False


def get_attributes(parent_schema: openapi.Schema, parent_class_name: str, module: ModulePath, resolver: ResolverFunc) -> list[AttributeModel]:
    def is_required(schema: openapi.Schema, prop_name: str) -> bool:
        return schema.required is not None and prop_name in schema.required

    return [
        get_attribute(prop_schema, name, parent_class_name, is_required(parent_schema, name), module, resolver)
        for name, prop_schema in parent_schema.properties.items()
    ]


def get_attribute(typ: SchemaOrRef, name: str, parent_name: str, required: bool, module: ModulePath, resolve: ResolverFunc) -> AttributeModel:
    return AttributeModel(
        name=name.replace(':', '_'),
        annotation=get_attr_annotation(typ, name, parent_name, required, module, resolve),
    )


def get_enum_attribute(value: Any) -> AttributeModel:
    name = _name_for_value(value)
    value = "'" + value.replace("'", r"\'") + "'" if value is not None else None
    return AttributeModel(
        name=name,
        annotation=AttributeAnnotationModel(
            type=BuiltinTypeRef.from_type(type(value)),
            field_props={'default': value},
        )
    )


def _name_for_value(value: Any) -> str:
    if value is None: return 'none'
    result = re.compile(r'\W+').sub('_', str(value))
    if result == '' or not result[0].isalpha():
        result = 'v_' + result
    if result in PYTHON_KEYWORDS:
        result += '_'
    return result



from dataclasses import dataclass, field

from typing import TypeVar, TypeAlias

from lapis.render.module_path import ModulePath

T = TypeVar('T')

SchemaImport: TypeAlias = tuple[str, str]

template_imports = ['typing', 'builtins']


@dataclass(frozen=True, kw_only=True)
class AbstractModule:
    path: ModulePath
    imports: list[str] = field(default_factory=list)
    type_checking_imports: list[SchemaImport] = field(default_factory=list)



import logging
import re
from dataclasses import dataclass
from typing import Optional, Union

import inflection

from .attribute import AttributeModel
from .attribute_annotation import AttributeAnnotationModel
from .refs import ResolverFunc
from ..module_path import ModulePath
from ..type_ref import TypeRef, get_type_ref
from ...openapi import model as openapi

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class OperationFunctionModel:
    name: str
    method: str
    path: str
    params: list[AttributeModel]
    params_model_name: Optional[TypeRef]
    result_class_map: dict[str, TypeRef]
    docstr: Optional[str] = None


_FIELD_PROPS = dict(
    in_=None,
    description=None,
    allowEmptyValue=False,
    style=None,
    explode=None,
    allowReserved=False,
)


def get_operation_param(param: Union[openapi.Parameter, openapi.Reference], parent_name: str, module: ModulePath, resolve: ResolverFunc) -> AttributeModel:
    if isinstance(param, openapi.Reference):
        param, module, _ = resolve(param, openapi.Parameter)

    schema = param.schema_
    if isinstance(schema, openapi.Reference):
        schema, module, schema_name = resolve(schema, openapi.Schema)
    else:
        schema_name = inflection.camelize(parent_name) + inflection.camelize(param.name)
        module = module / 'schemas'

    field_props = {k: (getattr(param, k) or _FIELD_PROPS[k]) for k in _FIELD_PROPS}
    param_name = param.in_[0] + '_' + sanitise_param_name(param.name)

    field_props['alias'] = param.name

    return AttributeModel(
        name=param_name,
        annotation=AttributeAnnotationModel(
            type=get_type_ref(schema, module, schema_name, param.required, resolve),
            field_props=field_props,
        ),
        deprecated=param.deprecated,
    )


def sanitise_param_name(param_name: str) -> str:
    # potential name clash
    return re.compile(r'\W+').sub('_', param_name)


def resolve_type_ref(typ: Union[openapi.Schema, openapi.Reference], module: ModulePath, name: str, resolver: ResolverFunc) -> TypeRef:
    if isinstance(typ, openapi.Reference):
        typ, module, name = resolver(typ, openapi.Schema)
    return get_type_ref(typ, module, name, True, resolver)


def get_operation_func(op: openapi.Operation, method: str, url_path: str, module: ModulePath, resolver: ResolverFunc) -> OperationFunctionModel:
    params = [get_operation_param(oapi_param, op.operationId, module, resolver) for oapi_param in op.parameters] if op.parameters else []

    result_class_map = {
        resp_code: resolve_type_ref(list(typ.content.values())[0].schema_, module, inflection.camelize(op.operationId) + resp_code + 'Response', resolver)
        for resp_code, typ in op.responses.__root__.items()
        if typ.content
    }

    return OperationFunctionModel(
        name=op.operationId,
        method=method,
        path=re.compile(r'\{([^}]+)\}').sub(r'{p_\1}', url_path),
        params=params,
        params_model_name=TypeRef(module=(module / 'schemas').str(), name=inflection.camelize(op.operationId)) if op.parameters else None,
        result_class_map=result_class_map,
    )
